# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/Gait_Analysis',
                'doc_host': 'https://Defense-Circuits-Lab.github.io',
                'git_url': 'https://github.com/Defense-Circuits-Lab/Gait_Analysis',
                'lib_path': 'gait_analysis'},
  'syms': { 'gait_analysis.core': { 'gait_analysis.core.Bodypart': ('core.html#bodypart', 'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart.__init__': ('core.html#bodypart.__init__', 'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart._add_immobility_to_df': ( 'core.html#bodypart._add_immobility_to_df',
                                                                                           'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart._add_rolling_speed_to_df': ( 'core.html#bodypart._add_rolling_speed_to_df',
                                                                                              'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart._add_speed_to_df': ( 'core.html#bodypart._add_speed_to_df',
                                                                                      'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart._apply_exclusion_criteria': ( 'core.html#bodypart._apply_exclusion_criteria',
                                                                                               'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart._slice_df': ('core.html#bodypart._slice_df', 'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart.calculate_speed_and_identify_immobility': ( 'core.html#bodypart.calculate_speed_and_identify_immobility',
                                                                                                             'gait_analysis/core.py'),
                                    'gait_analysis.core.Bodypart.exclusion_criteria': ( 'core.html#bodypart.exclusion_criteria',
                                                                                        'gait_analysis/core.py'),
                                    'gait_analysis.core.EventBout': ('core.html#eventbout', 'gait_analysis/core.py'),
                                    'gait_analysis.core.EventBout.__init__': ('core.html#eventbout.__init__', 'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording': ('core.html#trackedrecording', 'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.__init__': ( 'core.html#trackedrecording.__init__',
                                                                                      'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording._attempt_loading_df_from_dlc_output': ( 'core.html#trackedrecording._attempt_loading_df_from_dlc_output',
                                                                                                                 'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording._load_remaining_metadata': ( 'core.html#trackedrecording._load_remaining_metadata',
                                                                                                      'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.export_results': ( 'core.html#trackedrecording.export_results',
                                                                                            'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.inspect_processing': ( 'core.html#trackedrecording.inspect_processing',
                                                                                                'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.marker_ids_to_exclude_for_smoothing_and_interpolation': ( 'core.html#trackedrecording.marker_ids_to_exclude_for_smoothing_and_interpolation',
                                                                                                                                   'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.preprocess_tracking': ( 'core.html#trackedrecording.preprocess_tracking',
                                                                                                 'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.run_event_detection': ( 'core.html#trackedrecording.run_event_detection',
                                                                                                 'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.valid_mouse_lines': ( 'core.html#trackedrecording.valid_mouse_lines',
                                                                                               'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.valid_paradigms': ( 'core.html#trackedrecording.valid_paradigms',
                                                                                             'gait_analysis/core.py'),
                                    'gait_analysis.core.TrackedRecording.valid_recording_weeks': ( 'core.html#trackedrecording.valid_recording_weeks',
                                                                                                   'gait_analysis/core.py')},
            'gait_analysis.inspect': { 'gait_analysis.inspect.check_data_availability': ( 'inspect.html#check_data_availability',
                                                                                          'gait_analysis/inspect.py'),
                                       'gait_analysis.inspect.collect_all_available_data': ( 'inspect.html#collect_all_available_data',
                                                                                             'gait_analysis/inspect.py'),
                                       'gait_analysis.inspect.filter_dataframe': ( 'inspect.html#filter_dataframe',
                                                                                   'gait_analysis/inspect.py'),
                                       'gait_analysis.inspect.get_metadata_from_filename': ( 'inspect.html#get_metadata_from_filename',
                                                                                             'gait_analysis/inspect.py'),
                                       'gait_analysis.inspect.get_only_matching_xlsx_files': ( 'inspect.html#get_only_matching_xlsx_files',
                                                                                               'gait_analysis/inspect.py'),
                                       'gait_analysis.inspect.plot': ('inspect.html#plot', 'gait_analysis/inspect.py')},
            'gait_analysis.twoD.cam_interfaces': { 'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording': ( 'twod_cam_interfaces.html#toptracked2drecording',
                                                                                                                'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._add_to_logs': ( 'twod_cam_interfaces.html#toptracked2drecording._add_to_logs',
                                                                                                                             'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._check_metadata': ( 'twod_cam_interfaces.html#toptracked2drecording._check_metadata',
                                                                                                                                'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._get_correct_fps': ( 'twod_cam_interfaces.html#toptracked2drecording._get_correct_fps',
                                                                                                                                 'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._get_normalized_maze_corners': ( 'twod_cam_interfaces.html#toptracked2drecording._get_normalized_maze_corners',
                                                                                                                                             'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._get_parameters_to_normalize_maze_coordinates': ( 'twod_cam_interfaces.html#toptracked2drecording._get_parameters_to_normalize_maze_coordinates',
                                                                                                                                                              'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._load_remaining_metadata': ( 'twod_cam_interfaces.html#toptracked2drecording._load_remaining_metadata',
                                                                                                                                         'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._plot_selected_marker_ids_on_normalized_maze': ( 'twod_cam_interfaces.html#toptracked2drecording._plot_selected_marker_ids_on_normalized_maze',
                                                                                                                                                             'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording._retrieve_metadata': ( 'twod_cam_interfaces.html#toptracked2drecording._retrieve_metadata',
                                                                                                                                   'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording.export_results': ( 'twod_cam_interfaces.html#toptracked2drecording.export_results',
                                                                                                                               'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording.inspect_processing': ( 'twod_cam_interfaces.html#toptracked2drecording.inspect_processing',
                                                                                                                                   'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording.preprocess_tracking': ( 'twod_cam_interfaces.html#toptracked2drecording.preprocess_tracking',
                                                                                                                                    'gait_analysis/twoD/cam_interfaces.py'),
                                                   'gait_analysis.twoD.cam_interfaces.TopTracked2DRecording.run_event_detection': ( 'twod_cam_interfaces.html#toptracked2drecording.run_event_detection',
                                                                                                                                    'gait_analysis/twoD/cam_interfaces.py')},
            'gait_analysis.twoD.detect': { 'gait_analysis.twoD.detect.add_event_bouts_to_behavior_df': ( 'twod_detect.html#add_event_bouts_to_behavior_df',
                                                                                                         'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.add_immobility_based_on_several_bodyparts_to_behavior_df': ( 'twod_detect.html#add_immobility_based_on_several_bodyparts_to_behavior_df',
                                                                                                                                   'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.add_orientation_to_behavior_df': ( 'twod_detect.html#add_orientation_to_behavior_df',
                                                                                                         'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.create_behavior_df': ( 'twod_detect.html#create_behavior_df',
                                                                                             'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.create_event_objects': ( 'twod_detect.html#create_event_objects',
                                                                                               'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.get_gait_disruption_events': ( 'twod_detect.html#get_gait_disruption_events',
                                                                                                     'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.get_gait_events': ( 'twod_detect.html#get_gait_events',
                                                                                          'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.get_immobility_related_events': ( 'twod_detect.html#get_immobility_related_events',
                                                                                                        'gait_analysis/twoD/detect.py'),
                                           'gait_analysis.twoD.detect.get_interval_border_idxs_from_event_type_and_id': ( 'twod_detect.html#get_interval_border_idxs_from_event_type_and_id',
                                                                                                                          'gait_analysis/twoD/detect.py')},
            'gait_analysis.twoD.export': { 'gait_analysis.twoD.export.add_results_to_session_overview': ( 'twod_export.html#add_results_to_session_overview',
                                                                                                          'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.create_parameter_settings_df': ( 'twod_export.html#create_parameter_settings_df',
                                                                                                       'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.create_session_overview_df': ( 'twod_export.html#create_session_overview_df',
                                                                                                     'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.export_gait_related_bouts': ( 'twod_export.html#export_gait_related_bouts',
                                                                                                    'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.export_immobility_related_bouts': ( 'twod_export.html#export_immobility_related_bouts',
                                                                                                          'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_all_bout_ids': ( 'twod_export.html#get_all_bout_ids',
                                                                                           'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_bout_duration_per_bout_id': ( 'twod_export.html#get_bout_duration_per_bout_id',
                                                                                                        'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_bout_id_splits_depending_on_direction': ( 'twod_export.html#get_bout_id_splits_depending_on_direction',
                                                                                                                    'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_column_name_from_substring': ( 'twod_export.html#get_column_name_from_substring',
                                                                                                         'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_column_values_at_event_borders': ( 'twod_export.html#get_column_values_at_event_borders',
                                                                                                             'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_distance_covered_per_event': ( 'twod_export.html#get_distance_covered_per_event',
                                                                                                         'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.get_interval_start_and_end_time_per_event': ( 'twod_export.html#get_interval_start_and_end_time_per_event',
                                                                                                                    'gait_analysis/twoD/export.py'),
                                           'gait_analysis.twoD.export.write_xlsx_file_to_disk': ( 'twod_export.html#write_xlsx_file_to_disk',
                                                                                                  'gait_analysis/twoD/export.py')},
            'gait_analysis.twoD.preprocess': { 'gait_analysis.twoD.preprocess.add_new_marker_derived_from_existing_markers': ( 'twod_preprocess.html#add_new_marker_derived_from_existing_markers',
                                                                                                                               'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.compute_angle_error': ( 'twod_preprocess.html#compute_angle_error',
                                                                                                      'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.compute_distance_ratio_error': ( 'twod_preprocess.html#compute_distance_ratio_error',
                                                                                                               'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.compute_error_proportion': ( 'twod_preprocess.html#compute_error_proportion',
                                                                                                           'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.convert_df_to_cm': ( 'twod_preprocess.html#convert_df_to_cm',
                                                                                                   'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.create_bodypart_objects': ( 'twod_preprocess.html#create_bodypart_objects',
                                                                                                          'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.evaluate_maze_shape_using_open_corners': ( 'twod_preprocess.html#evaluate_maze_shape_using_open_corners',
                                                                                                                         'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_all_unique_marker_ids': ( 'twod_preprocess.html#get_all_unique_marker_ids',
                                                                                                            'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_conversion_factor_px_to_cm': ( 'twod_preprocess.html#get_conversion_factor_px_to_cm',
                                                                                                                 'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_corner_coords_with_likelihoods': ( 'twod_preprocess.html#get_corner_coords_with_likelihoods',
                                                                                                                     'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_distance_between_two_points': ( 'twod_preprocess.html#get_distance_between_two_points',
                                                                                                                  'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_low_likelihood_interval_border_idxs': ( 'twod_preprocess.html#get_low_likelihood_interval_border_idxs',
                                                                                                                          'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_most_reliable_marker_position_with_likelihood': ( 'twod_preprocess.html#get_most_reliable_marker_position_with_likelihood',
                                                                                                                                    'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_preprocessing_relevant_marker_ids': ( 'twod_preprocess.html#get_preprocessing_relevant_marker_ids',
                                                                                                                        'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_rotation_angle_with_closed_corners_only': ( 'twod_preprocess.html#get_rotation_angle_with_closed_corners_only',
                                                                                                                              'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_rotation_angle_with_open_corner': ( 'twod_preprocess.html#get_rotation_angle_with_open_corner',
                                                                                                                      'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.get_translation_vector': ( 'twod_preprocess.html#get_translation_vector',
                                                                                                         'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.interpolate_low_likelihood_intervals': ( 'twod_preprocess.html#interpolate_low_likelihood_intervals',
                                                                                                                       'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.normalize_df': ( 'twod_preprocess.html#normalize_df',
                                                                                               'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.rotate_df': ( 'twod_preprocess.html#rotate_df',
                                                                                            'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.smooth_tracked_coords_and_likelihood': ( 'twod_preprocess.html#smooth_tracked_coords_and_likelihood',
                                                                                                                       'gait_analysis/twoD/preprocess.py'),
                                               'gait_analysis.twoD.preprocess.translate_df': ( 'twod_preprocess.html#translate_df',
                                                                                               'gait_analysis/twoD/preprocess.py')},
            'gait_analysis.twoD.utils': { 'gait_analysis.twoD.utils.compute_coverage': ( 'twod_utils.html#compute_coverage',
                                                                                         'gait_analysis/twoD/utils.py'),
                                          'gait_analysis.twoD.utils.get_column_names': ( 'twod_utils.html#get_column_names',
                                                                                         'gait_analysis/twoD/utils.py'),
                                          'gait_analysis.twoD.utils.get_idxs_where_all_markers_exceed_likelihood': ( 'twod_utils.html#get_idxs_where_all_markers_exceed_likelihood',
                                                                                                                     'gait_analysis/twoD/utils.py'),
                                          'gait_analysis.twoD.utils.get_interval_border_idxs': ( 'twod_utils.html#get_interval_border_idxs',
                                                                                                 'gait_analysis/twoD/utils.py'),
                                          'gait_analysis.twoD.utils.get_max_odd_n_frames_for_time_interval': ( 'twod_utils.html#get_max_odd_n_frames_for_time_interval',
                                                                                                               'gait_analysis/twoD/utils.py'),
                                          'gait_analysis.twoD.utils.process_all_dlc_tracking_h5s_with_default_settings': ( 'twod_utils.html#process_all_dlc_tracking_h5s_with_default_settings',
                                                                                                                           'gait_analysis/twoD/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_twoD_00_utils.ipynb.

# %% auto 0
__all__ = ['process_all_dlc_tracking_h5s_with_default_settings', 'get_max_odd_n_frames_for_time_interval', 'get_column_names',
           'get_interval_border_idxs', 'compute_coverage', 'get_idxs_where_all_markers_exceed_likelihood']

# %% ../../nbs/01_twoD_00_utils.ipynb 3
from typing import List, Tuple, Dict, Optional
from pathlib import Path
import pandas as pd
import numpy as np
from tqdm import tqdm
from scipy.signal import savgol_filter

from ..core import EventBout
from . import cam_interfaces

# %% ../../nbs/01_twoD_00_utils.ipynb 4
# ToDo: If additional camera interfaces will be implemented, they should be specifieable here


def process_all_dlc_tracking_h5s_with_default_settings(
    in_dir_path: Path,  # path to the input directory which contains all DLC tracking data results
    out_dir_path: Path,  # path to the output directory where all processed results will be saved
    week_id: int,  # number of weeks post injection
) -> None:
    filepaths_dlc_trackings = []
    for filepath in in_dir_path.iterdir():
        if filepath.name.endswith(".h5"):
            if "filtered" not in filepath.name:
                filepaths_dlc_trackings.append(filepath)
    for filepath in tqdm(filepaths_dlc_trackings):
        recording = cam_interfaces.TopTracked2DRecording(
            filepath=filepath, week_id=week_id
        )
        if recording.df_successfully_loaded:
            recording.preprocess_tracking()
            if (
                recording.logs["coverage_critical_markers"]
                >= recording.logs["coverage_threshold"]
            ):
                recording.run_event_detection()
                recording.export_results(out_dir_path=out_dir_path)
                recording.inspect_processing()

# %% ../../nbs/01_twoD_00_utils.ipynb 5
def get_max_odd_n_frames_for_time_interval(
    fps: int,  # frames per second of the recording
    time_interval: 0.5,  # desired maximal time interval in seconds; default = 0.5 s
) -> int:
    """
    For the savgol_filter function of scipy - which will be used during preprocessing to smooth the data -
    you need an odd integer as the window_length parameter. This function helps to find the maximum odd number
    of frames that still fit within a specified time interval at a given fps.
    """
    assert type(fps) == int, '"fps" has to be an integer!'
    frames_per_time_interval = fps * time_interval
    if frames_per_time_interval % 2 == 0:
        max_odd_frame_count = frames_per_time_interval - 1
    elif frames_per_time_interval == int(frames_per_time_interval):
        max_odd_frame_count = frames_per_time_interval
    else:
        frames_per_time_interval = int(frames_per_time_interval)
        if frames_per_time_interval % 2 == 0:
            max_odd_frame_count = frames_per_time_interval - 1
        else:
            max_odd_frame_count = frames_per_time_interval
    assert (
        max_odd_frame_count > 0
    ), f"The specified time interval is too short to fit an odd number of frames"
    return int(max_odd_frame_count)

# %% ../../nbs/01_twoD_00_utils.ipynb 6
def get_column_names(
    df: pd.DataFrame,
    column_identifiers: List[str],
    marker_ids: List[str] = ["all"],
) -> List[str]:
    matching_column_names = []
    for column_name in df.columns:
        marker_id, column_identifier = column_name.split("_")
        if marker_ids == ["all"]:
            if column_identifier in column_identifiers:
                matching_column_names.append(column_name)
        else:
            if (marker_id in marker_ids) and (column_identifier in column_identifiers):
                matching_column_names.append(column_name)
    return matching_column_names

# %% ../../nbs/01_twoD_00_utils.ipynb 7
def get_interval_border_idxs(
    all_matching_idxs: np.ndarray,
    framerate: float,
    min_interval_duration: Optional[float] = None,
    max_interval_duration: Optional[float] = None,
) -> List[Tuple[int, int]]:
    interval_border_idxs = []
    if all_matching_idxs.shape[0] >= 1:
        step_idxs = np.where(np.diff(all_matching_idxs) > 1)[0]
        step_end_idxs = np.concatenate(
            [step_idxs, np.array([all_matching_idxs.shape[0] - 1])]
        )
        step_start_idxs = np.concatenate([np.array([0]), step_idxs + 1])
        interval_start_idxs = all_matching_idxs[step_start_idxs]
        interval_end_idxs = all_matching_idxs[step_end_idxs]
        for start_idx, end_idx in zip(interval_start_idxs, interval_end_idxs):
            interval_frame_count = (end_idx + 1) - start_idx
            interval_duration = interval_frame_count * framerate
            if (min_interval_duration != None) and (max_interval_duration != None):
                append_interval = (
                    min_interval_duration <= interval_duration <= max_interval_duration
                )
            elif min_interval_duration != None:
                append_interval = min_interval_duration <= interval_duration
            elif max_interval_duration != None:
                append_interval = interval_duration <= max_interval_duration
            else:
                append_interval = True
            if append_interval:
                interval_border_idxs.append((start_idx, end_idx))
    return interval_border_idxs

# %% ../../nbs/01_twoD_00_utils.ipynb 8
def compute_coverage(
    df: pd.DataFrame, critical_marker_ids: List[str], likelihood_threshold: float = 0.5
) -> float:
    idxs_where_all_markers_exceed_likelihood_threshold = (
        get_idxs_where_all_markers_exceed_likelihood(
            df=df,
            marker_ids=critical_marker_ids,
            likelihood_threshold=likelihood_threshold,
        )
    )
    return idxs_where_all_markers_exceed_likelihood_threshold.shape[0] / df.shape[0]

# %% ../../nbs/01_twoD_00_utils.ipynb 9
def get_idxs_where_all_markers_exceed_likelihood(
    df: pd.DataFrame, marker_ids: List[str], likelihood_threshold: float = 0.5
) -> np.ndarray:
    valid_idxs_per_marker_id = []
    for marker_id in marker_ids:
        valid_idxs_per_marker_id.append(
            df.loc[df[f"{marker_id}_likelihood"] >= likelihood_threshold].index.values
        )
    shared_valid_idxs_for_all_markers = valid_idxs_per_marker_id[0]
    if len(valid_idxs_per_marker_id) > 1:
        for i in range(1, len(valid_idxs_per_marker_id)):
            shared_valid_idxs_for_all_markers = np.intersect1d(
                shared_valid_idxs_for_all_markers, valid_idxs_per_marker_id[i]
            )
    return shared_valid_idxs_for_all_markers



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_twoD_04_export.ipynb.

# %% auto 0
__all__ = ['export_immobility_related_bouts', 'get_all_bout_ids', 'get_bout_duration_per_bout_id',
           'get_column_values_at_event_borders', 'get_distance_covered_per_event',
           'get_interval_start_and_end_time_per_event', 'export_gait_related_bouts', 'create_session_overview_df',
           'get_bout_id_splits_depending_on_direction', 'get_column_name_from_substring',
           'add_results_to_session_overview', 'create_parameter_settings_df', 'write_xlsx_file_to_disk']

# %% ../../nbs/01_twoD_04_export.ipynb 3
from typing import List, Tuple, Dict, Optional
from pathlib import Path
import pandas as pd
import numpy as np

# %% ../../nbs/01_twoD_04_export.ipynb 4
def export_immobility_related_bouts(
    df: pd.DataFrame, event_type: str, framerate: float
) -> pd.DataFrame:
    results_per_event = {
        "bout_id": [],
        "duration": [],
        "CenterOfGravity_x_at_bout_start": [],
        "towards_open_at_bout_start": [],
        "distance_covered_cm": [],
        "start_time": [],
        "end_time": [],
    }
    results_per_event["bout_id"] = get_all_bout_ids(df=df, event_type=event_type)
    if len(results_per_event["bout_id"]) >= 1:
        results_per_event["duration"] = get_bout_duration_per_bout_id(
            df=df, event_type=event_type, event_ids=results_per_event["bout_id"]
        )
        x_positions_center_of_gravity_at_interval_borders = (
            get_column_values_at_event_borders(
                df=df,
                event_type=event_type,
                event_ids=results_per_event["bout_id"],
                column_name="CenterOfGravity_x",
            )
        )
        results_per_event[
            "CenterOfGravity_x_at_bout_start"
        ] = x_positions_center_of_gravity_at_interval_borders[:, 0]
        direction_towards_open_at_interval_borders = get_column_values_at_event_borders(
            df=df,
            event_type=event_type,
            event_ids=results_per_event["bout_id"],
            column_name="facing_towards_open_end",
        )
        results_per_event[
            "towards_open_at_bout_start"
        ] = direction_towards_open_at_interval_borders[:, 0]
        results_per_event["distance_covered_cm"] = get_distance_covered_per_event(
            df=df,
            event_type=event_type,
            event_ids=results_per_event["bout_id"],
            marker_id="CenterOfGravity",
        )
        bout_start_and_end_time = get_interval_start_and_end_time_per_event(
            df=df,
            event_type=event_type,
            event_ids=results_per_event["bout_id"],
            framerate=framerate,
        )
        results_per_event["start_time"] = bout_start_and_end_time[:, 0]
        results_per_event["end_time"] = bout_start_and_end_time[:, 1]
    return pd.DataFrame(data=results_per_event)

# %% ../../nbs/01_twoD_04_export.ipynb 5
def get_all_bout_ids(df: pd.DataFrame, event_type: str) -> np.ndarray:
    return df[f"{event_type}_id"].dropna().unique()

# %% ../../nbs/01_twoD_04_export.ipynb 6
def get_bout_duration_per_bout_id(
    df: pd.DataFrame, event_type: str, event_ids: List[float]
) -> List[float]:
    durations = []
    for event_id in event_ids:
        durations.append(
            df.loc[df[f"{event_type}_id"] == event_id, f"{event_type}_duration"].iloc[0]
        )
    return durations

# %% ../../nbs/01_twoD_04_export.ipynb 7
def get_column_values_at_event_borders(
    df: pd.DataFrame, event_type: str, event_ids: List[float], column_name: str
) -> np.ndarray:
    values_at_interval_borders = []
    for event_id in event_ids:
        start_value = df.loc[df[f"{event_type}_id"] == event_id, column_name].iloc[0]
        end_value = df.loc[df[f"{event_type}_id"] == event_id, column_name].iloc[-1]
        values_at_interval_borders.append((start_value, end_value))
    return np.asarray(values_at_interval_borders)

# %% ../../nbs/01_twoD_04_export.ipynb 8
def get_distance_covered_per_event(
    df: pd.DataFrame, event_type: str, event_ids: List[float], marker_id: str
) -> List[float]:
    distances_per_event = []
    for event_id in event_ids:
        df_tmp = df.loc[df[f"{event_type}_id"] == event_id].copy()
        distances_per_event.append(
            (
                (
                    df_tmp[f"{marker_id}_x"].diff() ** 2
                    + df_tmp[f"{marker_id}_y"].diff() ** 2
                )
                ** 0.5
            )
            .cumsum()
            .iloc[-1]
        )
    return distances_per_event

# %% ../../nbs/01_twoD_04_export.ipynb 9
def get_interval_start_and_end_time_per_event(
    df: pd.DataFrame, event_type: str, event_ids: List[float], framerate: float
) -> np.ndarray:
    interval_border_idxs = []
    for event_id in event_ids:
        start_time, end_time = (
            df.loc[df[f"{event_type}_id"] == event_id].index.values[[0, -1]] * framerate
        )
        interval_border_idxs.append((start_time, end_time))
    return np.asarray(interval_border_idxs)

# %% ../../nbs/01_twoD_04_export.ipynb 10
def export_gait_related_bouts(
    df: pd.DataFrame, event_type: str, framerate: float
) -> pd.DataFrame:
    # ToDo: very similar to "export_immobility_related_bouts" - can they be combined to one generalized version?
    results_per_event = {
        "bout_id": [],
        "duration": [],
        "CenterOfGravity_x_at_bout_end": [],
        "towards_open_at_bout_end": [],
        "distance_covered_cm": [],
        "start_time": [],
        "end_time": [],
    }
    results_per_event["bout_id"] = get_all_bout_ids(df=df, event_type=event_type)
    if len(results_per_event["bout_id"]) >= 1:
        results_per_event["duration"] = get_bout_duration_per_bout_id(
            df=df, event_type=event_type, event_ids=results_per_event["bout_id"]
        )
        x_positions_center_of_gravity_at_interval_borders = (
            get_column_values_at_event_borders(
                df=df,
                event_type=event_type,
                event_ids=results_per_event["bout_id"],
                column_name="CenterOfGravity_x",
            )
        )
        results_per_event[
            "CenterOfGravity_x_at_bout_end"
        ] = x_positions_center_of_gravity_at_interval_borders[:, 1]
        direction_towards_open_at_interval_borders = get_column_values_at_event_borders(
            df=df,
            event_type=event_type,
            event_ids=results_per_event["bout_id"],
            column_name="facing_towards_open_end",
        )
        results_per_event[
            "towards_open_at_bout_end"
        ] = direction_towards_open_at_interval_borders[:, 1]
        results_per_event["distance_covered_cm"] = get_distance_covered_per_event(
            df=df,
            event_type=event_type,
            event_ids=results_per_event["bout_id"],
            marker_id="CenterOfGravity",
        )
        bout_start_and_end_time = get_interval_start_and_end_time_per_event(
            df=df,
            event_type=event_type,
            event_ids=results_per_event["bout_id"],
            framerate=framerate,
        )
        results_per_event["start_time"] = bout_start_and_end_time[:, 0]
        results_per_event["end_time"] = bout_start_and_end_time[:, 1]
    return pd.DataFrame(data=results_per_event)

# %% ../../nbs/01_twoD_04_export.ipynb 11
def create_session_overview_df(
    dfs_to_export_with_individual_bout_dfs: Dict[str, pd.DataFrame]
) -> pd.DataFrame:
    session_overview = {
        "bout_type": [],
        "total_bouts_count": [],
        "total_duration": [],
        "total_distance_covered": [],
        "mean_duration": [],
        "mean_distance_covered": [],
        "mean_CenterOfGravity_x": [],
    }
    for tab_name, df in dfs_to_export_with_individual_bout_dfs.items():
        bout_ids_split_depending_on_direction = (
            get_bout_id_splits_depending_on_direction(df=df)
        )
        for (
            split_id,
            relevant_bout_ids,
        ) in bout_ids_split_depending_on_direction.items():
            session_overview = add_results_to_session_overview(
                session_overview=session_overview,
                df=df,
                event_type=tab_name,
                event_prefix=split_id,
                bout_ids=relevant_bout_ids,
            )
    return pd.DataFrame(data=session_overview)

# %% ../../nbs/01_twoD_04_export.ipynb 12
def get_bout_id_splits_depending_on_direction(
    df: pd.DataFrame,
) -> Dict[str, List[float]]:
    towards_open_column_name = get_column_name_from_substring(
        all_columns=list(df.columns), substring="towards_open"
    )
    bout_ids_split_by_direction = {
        "all": list(df["bout_id"].unique()),
        "towards_open": list(
            df.loc[df[towards_open_column_name] == True, "bout_id"].unique()
        ),
        "towards_closed": list(
            df.loc[df[towards_open_column_name] != True, "bout_id"].unique()
        ),
    }
    return bout_ids_split_by_direction

# %% ../../nbs/01_twoD_04_export.ipynb 13
def get_column_name_from_substring(all_columns: List[str], substring: str) -> str:
    matching_column_names = [
        column_name for column_name in all_columns if substring in column_name
    ]
    assert (
        len(matching_column_names) == 1
    ), f"There should be exactly one match for {substring} - however, {len(matching_column_names)} were found: [{matching_column_names}]."
    return matching_column_names[0]

# %% ../../nbs/01_twoD_04_export.ipynb 14
def add_results_to_session_overview(
    session_overview: Dict,
    df: pd.DataFrame,
    event_type: str,
    event_prefix: str,
    bout_ids: List[float],
) -> Dict:
    session_overview["bout_type"].append(f"{event_prefix}_{event_type}")
    if len(bout_ids) > 0:
        session_overview["total_bouts_count"].append(len(bout_ids))
        session_overview["total_duration"].append(
            df.loc[df["bout_id"].isin(bout_ids), "duration"].cumsum().iloc[-1]
        )
        session_overview["total_distance_covered"].append(
            df.loc[df["bout_id"].isin(bout_ids), "distance_covered_cm"]
            .cumsum()
            .iloc[-1]
        )
        session_overview["mean_duration"].append(
            df.loc[df["bout_id"].isin(bout_ids), "duration"].mean()
        )
        session_overview["mean_distance_covered"].append(
            df.loc[df["bout_id"].isin(bout_ids), "distance_covered_cm"].mean()
        )
        center_of_gravity_x_column_name = get_column_name_from_substring(
            all_columns=list(df.columns), substring="CenterOfGravity_x"
        )
        session_overview["mean_CenterOfGravity_x"].append(
            df.loc[df["bout_id"].isin(bout_ids), center_of_gravity_x_column_name].mean()
        )
    else:
        session_overview["total_bouts_count"].append(0)
        session_overview["total_duration"].append(0)
        session_overview["total_distance_covered"].append(0)
        session_overview["mean_duration"].append(np.nan)
        session_overview["mean_distance_covered"].append(np.nan)
        session_overview["mean_CenterOfGravity_x"].append(np.nan)
    return session_overview

# %% ../../nbs/01_twoD_04_export.ipynb 15
def create_parameter_settings_df(logs: Dict) -> pd.DataFrame:
    logged_settings = {"parameter": [], "specified_value": []}
    for parameter, value in logs.items():
        logged_settings["parameter"].append(parameter)
        logged_settings["specified_value"].append(value)
    return pd.DataFrame(data=logged_settings)

# %% ../../nbs/01_twoD_04_export.ipynb 16
def write_xlsx_file_to_disk(
    base_output_filepath: Path, dfs_to_export: Dict[str, pd.DataFrame]
) -> None:
    writer = pd.ExcelWriter(f"{base_output_filepath}.xlsx", engine="xlsxwriter")
    for tab_name, df in dfs_to_export.items():
        df.to_excel(writer, sheet_name=tab_name)
    writer.save()



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math
from typing import Tuple, List, Dict, Optional, Union
from pathlib import Path
from abc import ABC, abstractmethod
import os
import imageio.v3 as iio
from scipy.spatial.transform import Rotation
from scipy import interpolate
from scipy.linalg import norm
from scipy.signal import find_peaks, peak_widths
from statistics import mean
import cv2
import pickle


class Recording3D(ABC):
    """
    Class for Analysing 3D-Position Data of mice in the OpeningTrack.
    
    Attributes:
        full_df_from_csv(pandas.DataFrame): the Dataframe containing all bodyparts with x, y, z-coordinates and reprojection error as returned by Anipose
        recorded_framerate(int): fps of the recording
        metadata(Dict): dictionary containing information read from the filename, such as animal_id, recording_date and Opening Track paradigm
    """
    
    def __init__(self, filepath: Path, recorded_framerate: int)->None:
        """
        Constructor for the Recording3D class.
        
        This function calls functions to get the Dataframe from the csv, that is given as filepath argument and to read metadata from the filename.
        
        Parameters:
            filepath(pathlib.Path): the filepath to the csv containing Anipose data
            recorded_framerate(int): fps of the recording
        """
        self.full_df_from_csv = self._get_df_from_csv(filepath = filepath)
        self.recorded_framerate = recorded_framerate
        self.metadata = self._retrieve_metadata(filepath = filepath)
        
        
    def _get_df_from_csv(self, filepath: Path)->pd.DataFrame:
        """
        Reads the Dataframe from the csv-file.
        
        Parameters:
            filepath(pathlib.Path): the path linked to the csv.file
        Returns:
            pandas.DataFrame: containing all bodyparts with x, y, z-coordinates and reprojection error as returned by Anipose
        """
        if not filepath.endswith('.csv'):
            raise ValueError('The Path you specified is not linking to a .csv-file!')
        return pd.read_csv(filepath)
    
    
    def _retrieve_metadata(self, filepath: Path)->Dict:
        """
        Function, that slices the Filename to get the encoded metadata.
        
        Relying on file naming like this: 196_F7-27_220826_OTT.csv
        
        Parameters:
            filepath(pathlib.Path): the path linked to the csv.file
        Returns:
            Dict: containing date of recording, animal_id and OT paradigm
        """
        filepath_slices = filepath.split('_')
        animal_line, animal_id, recording_date, paradigm = filepath_slices[0], filepath_slices[1], filepath_slices[2], filepath_slices[3]
        return {'recording_date': recording_date, 'animal': animal_line + '_' + animal_id, 'paradigm': paradigm}
        
        
    def run(self)->None:
        """
        Function to create Bodypart objects for all the tracked markers.
        
        A function is called, that first calculates the centerofgravity and then creates Bodypart objects for all the markers.
        The coordinate system is then normalized via translation, rotation and conversion to unit cms.
        Basic parameters for the bodyparts are already calculated, such as, speed and immobility.
        Currently no frames are excluded.
        """
        self._calculate_center_of_gravity()
        #todo: calculate error?
        self._create_all_bodyparts()
        self._normalize_coordinate_system()
        #add error to find best matching frame?
        self._run_basic_operations_on_bodyparts()
        self._get_tracking_stability()
        

    def _calculate_center_of_gravity(self)->None:
        """
        Function, that calculates the centerofgravity.
        
        The center_of_gravity is calculated using the bodyparts Snout and TailBase.
        It adds centerofgravity to self.full_df_from_csv.
        """
        for coordinate in ['x', 'y', 'z']:
            self.full_df_from_csv[f'centerofgravity_{coordinate}'] = (self.full_df_from_csv[f'Snout_{coordinate}'] + self.full_df_from_csv[f'TailBase_{coordinate}'])/2
        self.full_df_from_csv['centerofgravity_error'] = 0
            
        
    def _create_all_bodyparts(self)->None:
        """
        Function, that creates a Dictionary with all Bodypart objects.
        
        The dictionary uses the label given from Deeplabcut tracking as key for the Bodypart objects.
        It sets the dictionary as self.bodyparts.
        """
        self.bodyparts = {}
        for key in self.full_df_from_csv.keys():
            bodypart = key.split('_')[0]
            if bodypart not in self.bodyparts.keys() and bodypart not in set (['M', 'center', 'fnum']):
                self.bodyparts[bodypart] = Bodypart(bodypart_id = bodypart, df = self.full_df_from_csv)
                
    
    
    def _normalize_coordinate_system(self)->None:
        """
        This Function normalizes the coordinate system.
        
        The mazecorners of the best frame are used to calculate necessary parameters for the following functions, 
        such as the conversion factor from the intrinsic unit to cms, 
        the translation vector from the real-world-coordinate system to the null-space,
        and the rotation_matrix for 3D rotation, based on calculation of Euler-angles.
        With this parameters it calls the normalization for each single bodypart.
        """
        mazecorners = self._fix_coordinates_of_maze_corners()
        conversion_factor = self._get_conversion_factor_to_cm(reference_points = mazecorners)
        translation_vector = self._get_translation_vector(reference_points = mazecorners)
        rotation_matrix = self._get_rotation_matrix(reference_points = mazecorners)
        
        for bodypart in self.bodyparts.values():
            bodypart.normalize_df(translation_vector = translation_vector, rotation_matrix = rotation_matrix, conversion_factor = conversion_factor)
    
    def _find_best_mazecorners_for_normalization(self)->int:
        """
        Function to find the frame, where the mazecorners are tracked best.
        
        It creates Angle objects for the four corners and calculates the length of the long maze sides.
        This parameters are used to create a weighted residual mean of squares function to find the best matching frame.
        https://en.wikipedia.org/wiki/Residual_sum_of_squares
        
        Returns:
            best_matching_frame(int): the index of the frame
        """
        angle_open_right = Angle(bodypart_a = self.bodyparts['MazeCornerOpenRight'], bodypart_b = self.bodyparts['MazeCornerClosedRight'], object_to_calculate_angle = self.bodyparts['MazeCornerOpenLeft'])
        angle_open_left = Angle(bodypart_a = self.bodyparts['MazeCornerOpenLeft'], bodypart_b = self.bodyparts['MazeCornerClosedLeft'], object_to_calculate_angle = self.bodyparts['MazeCornerOpenRight'])
        angle_closed_right = Angle(bodypart_a = self.bodyparts['MazeCornerClosedRight'], bodypart_b = self.bodyparts['MazeCornerOpenRight'], object_to_calculate_angle = self.bodyparts['MazeCornerClosedLeft'])    
        angle_closed_left = Angle(bodypart_a = self.bodyparts['MazeCornerClosedLeft'], bodypart_b = self.bodyparts['MazeCornerOpenLeft'], object_to_calculate_angle = self.bodyparts['MazeCornerClosedRight'])
        
        length_right_side = np.sqrt(np.sum(self.bodyparts['MazeCornerOpenRight'].df_raw.loc[:, ('x', 'y', 'z')].values - self.bodyparts['MazeCornerClosedRight'].df_raw.loc[:, ('x', 'y', 'z')].values, axis=1)**2)
        length_left_side = np.sqrt(np.sum(self.bodyparts['MazeCornerOpenLeft'].df_raw.loc[:, ('x', 'y', 'z')].values - self.bodyparts['MazeCornerClosedLeft'].df_raw.loc[:, ('x', 'y', 'z')].values, axis=1)**2)
        
        w_angle = 1
        w_sides = 4
        #w_error = 4, use reprojection error in formula?
        frame_errors = []
        n = 4*w_angle + w_sides
        for frame in range(self.full_df_from_csv.shape[0]):
            weighted_residual_mean_of_squares = math.sqrt((w_angle*(angle_open_right.parameter_array[frame]-90)**2
                                                               +w_angle*(angle_open_left.parameter_array[frame]-90)**2
                                                               +w_angle*(angle_closed_right.parameter_array[frame]-90)**2
                                                               +w_angle*(angle_closed_left.parameter_array[frame]-90)**2
                                                               +w_sides*(length_right_side[frame]-length_left_side[frame])**2)/n)
            frame_errors.append(weighted_residual_mean_of_squares)
            # alle Fehler normieren? z.B. ((angle-90)/90)**2, ((length_right_side-length_left_side)/length_right_side)**2
        """
        weighted_residual_mean_of_squares = np.sqrt((w_angle*(angle_open_right.parameter_array-90)**2
                                                               +w_angle*(angle_open_left.parameter_array-90)**2
                                                               +w_angle*(angle_closed_right.parameter_array-90)**2
                                                               +w_angle*(angle_closed_left.parameter_array-90)**2
                                                               +w_sides*(length_right_side[frame]-length_left_side)**2)/n)
        """
        best_matching_frame = frame_errors.index(min(frame_errors))
        #print(angle_closed_left.parameter_array[best_matching_frame], angle_closed_right.parameter_array[best_matching_frame], angle_open_left.parameter_array[best_matching_frame], angle_open_right.parameter_array[best_matching_frame])
        return best_matching_frame
    
    def _get_conversion_factor_to_cm(self, reference_points: Tuple[np.array, np.array, np.array])->float:
        """
        Function to get the conversion factor of the unspecified unit to cm.
        
        Parameters:
            reference_points (Tuple): containing the vecotrs as np.arrays of the best tracked mazecorners
            
        Returns:
            conversion_factor(float): factor to convert the unspecified unit into cm.
        """
        conversion_factor = (50/np.sqrt(sum((reference_points[1]-reference_points[0])**2)))
        return conversion_factor
    
    def _get_translation_vector(self, reference_points:  Tuple[np.array, np.array, np.array])->np.array:
        """
        Function that calculates the offset of the real-world-coordinate system to the null space.
        
        Parameters:
            reference_points (Tuple): containing the vecotrs as np.arrays of the best tracked mazecorners
            
        Returns:
            translation_vector(np.array): vector with offset in each dimension
        """
        translation_vector = -reference_points[0]
        return translation_vector
    
    def _get_rotation_matrix(self, reference_points: Tuple[np.array, np.array, np.array])->Rotation:
        """
        Function that calculates the Rotation matrix.
        
        It calculates the Euler angles between the real-world-coordinate system (xyz) to the null space (XYZ) first,
        and inserts them into the rotation matrix.
        https://en.wikipedia.org/wiki/Euler_angles
        
        Parameters:
            reference_points (Tuple): containing the vecotrs as np.arrays of the best tracked mazecorners
        
        Returns:
            scipy.spatial.transform.Rotation: Rotation matrix from the three calculated angles
        """
        # The axes of the real-world-coordinate system are calculated
        x_axis, y_axis, z_axis = self._get_axes(reference_points=reference_points)
        # The angle between the x_axis and the intersection of the z_axis and the Z_axis are calculated.
        x_angle = self._calculate_angle(vector1=x_axis, vector2 = np.cross(z_axis, np.array([0, 0, 1])))
        # The xyz-system is rotated by that angle.
        reference_points_z = self._apply_single_rotation(angle = x_angle, reference_points=reference_points, axis = 'z')
        
        # The axes of the rotated coordinate system are calculated
        x_axis2, y_axis2, z_axis2 = self._get_axes(reference_points=reference_points_z)
        # The angle between the rotated z_axis and the Z_axis are calculated.
        z_angle = self._calculate_angle(vector1=z_axis2, vector2 = np.array([0, 0, 1]))
        # The xyz-system is rotated by that angle.
        reference_points_zx = self._apply_single_rotation(angle = z_angle, reference_points=reference_points_z, axis = 'x')

        # The axes of the rotated coordinate system are calculated
        x_axis3, y_axis3, z_axis3 = self._get_axes(reference_points=reference_points_zx)
        # The angle between the intersection of the z_axis and the Z_axis and the x_axis are calculated.
        x_angle2 = self._calculate_angle(vector1=np.cross(z_axis, np.array([0, 0, 1])), vector2 = np.array([1, 0, 0]))
        
        return Rotation.from_euler('zxz', [x_angle, z_angle, x_angle2], degrees=True)
   
    def _calculate_angle(self, vector1:np.array, vector2: np.array)->float:
        """
        Function to calculate the angle between two vectors.
        
        Parameters:
            vector1(np.array): first vector
            vector2(np.array): second vector
            
        Returns:
            float: the calculated angle in degrees
        """
        angle = np.dot(vector1, vector2)/(np.linalg.norm(vector1)*np.linalg.norm(vector2))
        return math.degrees(math.acos(angle))
    
    def _apply_single_rotation(self, angle: float, reference_points: Tuple[np.array, np.array, np.array], axis: str)->Tuple[np.array, np.array, np.array]:
        """
        Function that rotates the reference points around a given axis by a specified angle.
        
        It calculates the rotation matrix and applies it to the reference points.
        
        Parameters:
            angle(float): rotation angle 
            reference_points (Tuple): containing the vecotrs as np.arrays of the best tracked mazecorners
            axis: specifies the axis on which the rotation is performed
            
        Returns:
            Tuple: rotated reference_points
        
        """
        r = Rotation.from_euler(axis, [-angle], degrees=True)
        reference_points_rotated = []
        for point in reference_points:
             reference_points_rotated.append(r.apply(point, inverse=True)[0])
        return reference_points_rotated
    
    def _fix_coordinates_of_maze_corners(self)->Tuple[np.array, np.array, np.array]:
        """
        Function that creates the reference_points from the mazecorners.
        
        After finding the best matching frame it combines the -x, -y, -z coordinate of this frame for all corners into a np.array.
        
        Returns:
            reference_points(Tuple): the best tracked maze corners as numpy.Array
        """
        frame = self._find_best_mazecorners_for_normalization()
        
        mazecorneropenright = np.array([self.bodyparts['MazeCornerOpenRight'].df_raw.loc[frame, 'x'], self.bodyparts['MazeCornerOpenRight'].df_raw.loc[frame, 'y'], self.bodyparts['MazeCornerOpenRight'].df_raw.loc[frame, 'z']])
        mazecornerclosedright = np.array([self.bodyparts['MazeCornerClosedRight'].df_raw.loc[frame, 'x'], self.bodyparts['MazeCornerClosedRight'].df_raw.loc[frame, 'y'], self.bodyparts['MazeCornerClosedRight'].df_raw.loc[frame, 'z']])
        mazecorneropenleft = np.array([self.bodyparts['MazeCornerOpenLeft'].df_raw.loc[frame, 'x'], self.bodyparts['MazeCornerOpenLeft'].df_raw.loc[frame, 'y'], self.bodyparts['MazeCornerOpenLeft'].df_raw.loc[frame, 'z']])
        mazecornerclosedleft = np.array([self.bodyparts['MazeCornerClosedLeft'].df_raw.loc[frame, 'x'], self.bodyparts['MazeCornerClosedLeft'].df_raw.loc[frame, 'y'], self.bodyparts['MazeCornerClosedLeft'].df_raw.loc[frame, 'z']])
        return mazecornerclosedright, mazecorneropenright, mazecornerclosedleft

    def _get_axes(self, reference_points: Tuple[np.array, np.array, np.array])->Tuple[np.array, np.array, np.array]:
        """
        Calculates the axis-vectors from given reference points.
        
        Parameters:
            reference_points (Tuple): containing the vecotrs as np.arrays of the best tracked mazecorners
        Returns:
            Tuple: the calculated axes
        """
        x_axis = reference_points[0]-reference_points[1]
        y_axis = reference_points[2]-reference_points[0]
        # the cross product returns an vector, that is orthogonal on both of the factorized vectors, which is the z-axis in our case.
        z_axis = np.cross(x_axis, y_axis)
        # Could point in the negative direction as well!
        return x_axis, y_axis, z_axis
                
        
        
    def _run_basic_operations_on_bodyparts(self)->None:
        """ Basic parameters for the bodyparts are calculated, such as, speed and immobility. """
        for bodypart in self.bodyparts.values():
            bodypart.run_basic_operations(recorded_framerate = self.recorded_framerate)            
            
    def _get_tracking_stability(self)->None:
        """
        Function, that calculates the percentage of frames, in which a marker was detected.
        
        It sets self.tracking_stability as pandas.DataFrame with columns for all markers and the percentage value as first row.
        """
        tracking_dict = {bodypart.id: bodypart.check_tracking_stability() for bodypart in self.bodyparts.values()}
        #calculate standard derivation for fixed markers
        #for bodypart in set(['MazeCornerOpenLeft', 'MazeCornerOpenRight', 'MazeCornerClosedLeft', 'MazeCornerClosedRight', 'LED5']): 
            #tracking_dict['standard_derivation']=np.stddev([])
        self.tracking_stability = pd.DataFrame([tracking_dict], index=['over_total_session', 'over_gait_events'])
        
    
    def get_freezing_bouts(self)->None:
        """
        Function for the detection of freezing bouts.
        
        After calculation of important parameters such as direction, turns and immobility of the most relevant bodyparts,
        the freezing bouts are collected.
        """
        self._get_direction()
        self._get_turns()
        self._check_immobility_of_all_freezing_bodyparts()        
        # create class for each parameter and abstract parent class?
        self._get_immobility_bouts()
        # think of better definition of freezing/immobility
        self._run_operations_on_immobility_bouts()
        self._collect_freezing_bouts()

        
    def _get_direction(self)->None:
        """
        Checks frame by frame, whether the Snout is closer to the open end of the maze than the Ears.
        
        This is set as a parameter (in the future: Parameter object) self.facing_towards_open_end with Boolean values.
        """
        self.facing_towards_open_end = self._initialize_new_parameter(dtype=bool)
        self.facing_towards_open_end.loc[(self.bodyparts['Snout'].df.loc[:, 'x']>self.bodyparts['EarLeft'].df.loc[:, 'x']) &
                                    (self.bodyparts['Snout'].df.loc[:, 'x']>self.bodyparts['EarRight'].df.loc[:, 'x'])] = True
        
        
    def _get_turns(self)->None:
        """
        Function that checks for turning events.
        
        Based on self.facing_towards_open_end, the indices of events, where a mouse turns are extracted and the attributes
        self.turns_to_closed and self.turns_to_open are created as a list of EventBouts.
        """
        turn_indices = self.facing_towards_open_end.where(self.facing_towards_open_end.diff()==True).dropna().index
        self.turns_to_closed=[EventBout(start_index = start_index) for start_index in turn_indices if self.facing_towards_open_end[start_index-1]==True]
        self.turns_to_open=[EventBout(start_index = start_index) for start_index in turn_indices if self.facing_towards_open_end[start_index-1]==False]
        for turning_bout in self.turns_to_closed:
            turning_bout.get_position(centerofgravity=self.bodyparts["centerofgravity"])
        for turning_bout in self.turns_to_open:
            turning_bout.get_position(centerofgravity=self.bodyparts["centerofgravity"])
            
        
    def _initialize_new_parameter(self, dtype: type)->pd.Series:
        """   
        Creates a Series object for initializing a parameter.
        
        This Function will be replaced by the object Parameter in the future.
        
        Parameters:
            dtype(type): the default type of the new created parameter.
        Returns:
            pd.Series of an array in shape of n_frames with default values set to 0 (if dtype bool->False)
        """
        return pd.Series(np.zeros_like(np.arange(self.full_df_from_csv.shape[0]), dtype = dtype))
    
                    
    def _check_immobility_of_all_freezing_bodyparts(self)->None:     
        """
        Function, that checks frame by frame, whether the relevant bodyparts for freezing are immobile.
        
        The information is stored as attribute self.all_freezing_bodyparts_immobile.
        """
        self.all_freezing_bodyparts_immobile = self._initialize_new_parameter(dtype=bool)
        self.all_freezing_bodyparts_immobile.loc[(self.bodyparts['Snout'].df.loc[:, 'immobility']) & (self.bodyparts['TailBase'].df.loc[:, 'immobility'])] = True
        
        
    def _get_immobility_bouts(self)->None:
        """
        Function, that creates immobility EventBouts.
        
        The Function detects start and end of an immobility episode and creates EventBouts for every episode.
        Sets a List of EventBouts as attribute self.immobility_bouts.
        
        """
        #since the first frame is allways immobile, the first change in all freezing bodyparts can be seen as the start of a immobility episode
        #this function finds all immobility_bouts, checks for directionality and whether the freezing threshold was reached
        changes_from_immobility_to_mobility = self.all_freezing_bodyparts_immobile.where(self.all_freezing_bodyparts_immobile.diff()==True).dropna()
        start_indices_of_immobility_bouts = changes_from_immobility_to_mobility[::2]
        end_indices_of_immobility_bouts = changes_from_immobility_to_mobility[1::2]
        self.immobility_bouts = []
        for i in range(len(start_indices_of_immobility_bouts)):
            start_index, end_index = start_indices_of_immobility_bouts.index[i], end_indices_of_immobility_bouts.index[i]
            self.immobility_bouts.append(EventBout(start_index = start_index, end_index = end_index))
            
    def _run_operations_on_immobility_bouts(self)->None:
        """
        Basic operations are run on the Immobility Bouts.
        
        A pandas.DataFrame for immobility bouts is created and set as self.immobility_bout_df.
        """
        for immobility_bout in self.immobility_bouts:
            immobility_bout.check_direction(facing_towards_open_end=self.facing_towards_open_end)
            immobility_bout.check_that_freezing_threshold_was_reached(recorded_framerate=self.recorded_framerate)
            immobility_bout.get_position(centerofgravity=self.bodyparts["centerofgravity"])
        self.immobility_bout_df = pd.DataFrame([immobility_bout.dict for immobility_bout in self.immobility_bouts])
            
            
    def _collect_freezing_bouts(self)->None:
        """
        The Immobility bouts, where the freezing threshold was exceeded are collected as freezing bouts.
        
        A pandas.DataFrame for freezing bouts is created and set as self.freezing_bout_df.
        """
        self.freezing_bouts = []
        for immobility_bout in self.immobility_bouts:
            if immobility_bout.freezing_threshold_reached:
                self.freezing_bouts.append(immobility_bout)
        self.freezing_bout_df = pd.DataFrame([freezing_bout.dict for freezing_bout in self.freezing_bouts])
         
    def run_gait_analysis(self)->None:
        """
        Function, that runs functions, necessary for gait analysis.
        
        Angles between bodyparts of interest are calculated as Angle objects.
        A peak detection algorithm on paw_speed is used to detect steps.
        EventBouts are created.
        """
        steps = self._detect_steps()
        gait_periodes = self._define_gait_periodes(steps=steps)
        self._get_gait_event_bouts(gait_events=gait_periodes)
        self._get_tracking_stability_in_gait_periodes(gait_periodes=gait_periodes)
        self._calculate_parameters_for_gait_analysis()
        self._calculate_angles()
        self._add_parameters_to_steps(gait_events=gait_periodes)        
        
    def _detect_steps(self)->List:
        """
        Function that runs step detection in the individual paw Bodypart objects and sorts the steps by start_index.
        
        Returns:
            steps(List): Step objects for speed peaks in all paws
        """
        steps_per_paw = [self.bodyparts[paw]._detect_steps() for paw in['HindPawRight', 'HindPawLeft', 'ForePawRight', 'ForePawLeft']]
        steps = [item for sublist in steps_per_paw for item in sublist]
        steps_sorted = sorted(steps, key=lambda x: getattr(x, 'start_index'))
        return steps_sorted
    
    def _define_gait_periodes(self, steps: List)->List:
        """
        Function, that recognizes periodes of gait.
        
        If more than 3 steps are done in less than 3 seconds, those steps are considered a gait_event.

        Parameters:
            steps(List): list of steps with Step objects
        Returns:
            gait_events(List): nested list with sublists that represent single gait events containing single step indices
        """
        gait_events = []
        for i in range(len(steps)):
            if i == 0:
                gait_event = []
            gait_event.append(steps[i])
            if i == len(steps)-1:
                if len(gait_event)>3:
                    gait_events.append(gait_event)
            elif (steps[i+1].start_index - steps[i].start_index) > 3*self.recorded_framerate:
                # 3s, where the mouse didn't do a step
                # create new gait event
                if len(gait_event)>3:
                    gait_events.append(gait_event)
                gait_event = []
            # create EventBouts for Gait?
        return gait_events
    
    def _get_gait_event_bouts(self, gait_events: List)->None:
        """
        Function that collects eventbouts of a specific type after a gait_event together.
        
        It sets the attribute self.turns_to_closed_after_gait for turns to the closed side after a gait event, 
        self.turns_to_open_after_gait for turns to the open side after a gait event,
        self.gait_disruption_bouts for immobility bouts after a gait event and
        self.freezing_of_gait_events for freezing bouts after a gait event.
        The lists are cleared from None values afterwards.
        
        Parameters:
            gait_events(List): nested list with sublists that represent single gait events containing single step indices
        """
        self.turns_to_closed_after_gait = [turn for turn in [self._bout_after_gait_event(gait_event = gait_event, event = self.turns_to_closed) for gait_event in gait_events] if type(turn)==EventBout]
        self.turns_to_open_after_gait = [turn for turn in [self._bout_after_gait_event(gait_event = gait_event, event = self.turns_to_open) for gait_event in gait_events] if type(turn)==EventBout]
        self.gait_disruption_bouts = [disruption for disruption in [self._bout_after_gait_event(gait_event = gait_event, event = self.immobility_bouts) for gait_event in gait_events] if type(disruption)==EventBout]
        self.freezing_of_gait_events = [freezing for freezing in [self._bout_after_gait_event(gait_event = gait_event, event = self.freezing_bouts) for gait_event in gait_events] if type(freezing)==EventBout]

            
    def _bout_after_gait_event(self, gait_event: List, event: List)->Union:
        """
        Function that checks the in and after a gait_event for an specified event.
        
        Parameters:
            frame_index(int): last index of a gait_event
            event(List): List with Event Bouts to check for
        Returns:
            Union: the Eventbout is returned if found, otherwise None
        """
        for bout in event:
            if bout.start_index in range(gait_event[0].start_index, gait_event[-1].start_index+self.recorded_framerate):
                return bout
                break
            else:
                pass
            
        
    def _calculate_angles(self)->None:
        """
        creates angles objects that are interesting for gait analysis
        """
        self.angle_hindkneeleft = Angle(bodypart_a = self.bodyparts['HindKneeLeft'], bodypart_b = self.bodyparts['BackAnkleLeft'], object_to_calculate_angle=self.bodyparts['HipLeft'])
        self.angle_backankleleft = Angle(bodypart_a = self.bodyparts['BackAnkleLeft'], bodypart_b = self.bodyparts['HindPawLeft'], object_to_calculate_angle=self.bodyparts['HindKneeLeft'])
        self.angle_hipleft = Angle(bodypart_a = self.bodyparts['HipLeft'], bodypart_b = self.bodyparts['HindKneeLeft'], object_to_calculate_angle=self.bodyparts['IliacCrestLeft'])
        
        self.angle_hindkneeright = Angle(bodypart_a = self.bodyparts['HindKneeRight'], bodypart_b = self.bodyparts['BackAngleRight'], object_to_calculate_angle=self.bodyparts['HipRight'])
        self.angle_backankleright = Angle(bodypart_a = self.bodyparts['BackAngleRight'], bodypart_b = self.bodyparts['HindPawRight'], object_to_calculate_angle=self.bodyparts['HindKneeRight'])
        self.angle_hipright = Angle(bodypart_a = self.bodyparts['HipRight'], bodypart_b = self.bodyparts['HindKneeRight'], object_to_calculate_angle=self.bodyparts['IliacCrestRight'])
        
        self.angle_wristleft = Angle(bodypart_a = self.bodyparts['WristLeft'], bodypart_b = self.bodyparts['ForePawLeft'], object_to_calculate_angle=self.bodyparts['ElbowLeft'])
        self.angle_elbowleft = Angle(bodypart_a = self.bodyparts['ElbowLeft'], bodypart_b = self.bodyparts['ShoulderLeft'], object_to_calculate_angle=self.bodyparts['WristLeft'])
        
        self.angle_wristright = Angle(bodypart_a = self.bodyparts['WristRight'], bodypart_b = self.bodyparts['ForePawRight'], object_to_calculate_angle=self.bodyparts['ElbowRight'])
        self.angle_elbowright = Angle(bodypart_a = self.bodyparts['ElbowRight'], bodypart_b = self.bodyparts['ShoulderRight'], object_to_calculate_angle=self.bodyparts['WristRight'])
    
    def _add_parameters_to_steps(self, gait_events: List)->None:
        gait_event = gait_events[2]
        for parameter in [self.angle_hindkneeleft.parameter_array, self.angle_hindkneeright.parameter_array, self.angle_backankleleft.parameter_array, self.angle_backankleright.parameter_array, self.angle_hipleft.parameter_array, self.angle_hipright.parameter_array, self.angle_wristleft.parameter_array, self.angle_wristright.parameter_array, self.angle_elbowleft.parameter_array, self.angle_elbowright.parameter_array, self.hind_stance, self.fore_stance]:
            plt.close()
            fig = plt.figure()
            plt.plot(parameter[gait_event[0].start_index:gait_event[-1].start_index])
            plt.show()
        pass
    
    def _calculate_parameters_for_gait_analysis(self)->None:        
        self.hind_stance_right = Stance(paw=self.bodyparts['HindPawRight'], object_a=self.bodyparts['centerofgravity'], object_b=self.bodyparts['TailBase'])
        self.hind_stance_left = Stance(paw=self.bodyparts['HindPawLeft'], object_a=self.bodyparts['centerofgravity'], object_b=self.bodyparts['TailBase'])
        self.fore_stance_right = Stance(paw=self.bodyparts['ForePawRight'], object_a=self.bodyparts['centerofgravity'], object_b=self.bodyparts['Snout'])
        self.fore_stance_left = Stance(paw=self.bodyparts['ForePawLeft'], object_a=self.bodyparts['centerofgravity'], object_b=self.bodyparts['Snout'])
        
        self.hind_stance = self.hind_stance_right.parameter_array + self.hind_stance_left.parameter_array
        self.fore_stance = self.fore_stance_right.parameter_array + self.fore_stance_left.parameter_array
        
        """
        Step Length
        Stride Length
                
        Gait Symmetry
        
        Angle Paw-BackAnkle/Wrist - Bodyaxis(Snout/TailBase/CoG)
        """
        pass
    
    def _create_PSTHs(self)->None:
        pass
    
    def _get_tracking_stability_in_gait_periodes(self, gait_periodes: List)->None:
        """
        Function, that calculates the percentage, in which a marker was detected during all gait periodes.
        
        This should reflect the quality of marker detection better than the whole session percentage.
        
        Parameters:
            gait_periods(List): nested list with sublists that represent single gait events containing single step indices
        """
        tracking_dict = dict.fromkeys(self.bodyparts.keys())
        for bodypart in self.bodyparts.values():
            tracking_dict[bodypart.id]=[]
            for gait_event in gait_periodes:
                tracking_dict[bodypart.id].append(bodypart.check_tracking_stability(start_end_index=(gait_event[0].start_index, gait_event[-1].start_index)))
            tracking_dict[bodypart.id]=mean(tracking_dict[bodypart.id])
        self.tracking_stability.loc['over_gait_events', :] = tracking_dict

                
class Step():
    def __init__(self, paw: str, start_index: int)->None:
        self.start_index=start_index
        self.paw=paw
                
    def _calculate_end_index(self)->None:
        pass
    
                                               
class Stance():
    """
    Class for calculation of the given paw to the bodyaxis as defined by object_a and object_b.
    
    Attributes:
        self.paw (Bodypart): Bodypart representation of the paw
        self.object_a: object, which defines the Bodyaxis
        self.object_b: object, which defines the Bodyaxis
        self.parameter_array: array, that contains the calculated distance (Stance) for every frame
    """
    def __init__(self, paw: 'Bodypart', object_a: 'Bodypart', object_b: 'Bodypart')->None:
        """
        Constructor for class Stance. It calls the function to calculate the stance already.
        
        The Stance is stored in self.parameter_array
        
        Parameters:
            self.paw (Bodypart): Bodypart representation of the paw
            self.object_a: object, which defines the Bodyaxis
            self.object_b: object, which defines the Bodyaxis
        """
        self.paw = paw
        self.object_a = object_a
        self.object_b = object_b
        self.parameter_array = self._shortest_distance_between_point_and_line()
    
    def _shortest_distance_between_point_and_line(self)->float:
        """
        Function to calculate the 
        
        First, the vector between the paw and one point on the bodyaxis is creeated as m0m1. Next, the direction vector of the line is defined.
        Last step: the normalized cross product between m0m1 and the bodyaxis vector divided by the normalized bodyaxis vector results in the shortest distance between the paw and the bodyaxis. https://onlinemschool.com/math/library/analytic_geometry/p_line/
        
        Returns:
            distance(float): shortest distance between the given paw and the bodyaxis.
        """
        m0m1 = np.array([self.paw.df['x'] - self.object_a.df['x'], self.paw.df['y'] - self.object_a.df['y'], self.paw.df['z'] - self.object_a.df['z']]).reshape(self.paw.df.shape[0], 3)
        line = np.array([self.object_a.df['x']-self.object_b.df['x'], self.object_a.df['y']-self.object_b.df['y'], self.object_a.df['z']-self.object_b.df['z']]).reshape(self.paw.df.shape[0], 3)
        distance = norm(np.cross(m0m1, line), axis=1, check_finite = False) / norm(line, axis=1, check_finite = False)
        return distance
    
        
class Bodypart():
    """
    Class that contains information for one single Bodypart.
    
    Attributes:
        self.id(str): Deeplabcut label of the bodypart
    """
    def __init__(self, bodypart_id: str, df: pd.DataFrame)->None:
        """
        Constructor for class Bodypart
        """
        self.id = bodypart_id
        self._get_sliced_df(df = df)
        
        
    def _get_sliced_df(self, df: pd.DataFrame)->None:
        """
        Function, that extracts the coordinates of a single bodypart.
        
        Parameters:
            df(pandas.DataFrame): the full dataframe of the recording with all bodyparts
        """
        self.df_raw = pd.DataFrame(data={'x': df.loc[:, self.id + '_x'], 'y': df.loc[:, self.id + '_y'], 'z': df.loc[:, self.id + '_z'], 'error': df.loc[:, self.id + '_error']})
    
        
    def normalize_df(self, translation_vector: np.array, rotation_matrix: Rotation, conversion_factor: float)->None:
        """
        Given the parameters, this function aligns the xyz-coordinate system with the null space.
        
        After translation to zero, rotation around the given angles and axes is performed and the units are converted into cm.
        The normalized dataframe is set as attribut self.df.
        
        Parameter:
            translation_vector(np.Array): vector with offset of xyz to XYZ in each dimension
            rotation_matrix(scipy.spatial.transform.Rotation): Rotation matrix obtained from Euler angles
            conversion_factor(float): factor to convert the unspecified unit into cm.
        """
        translated_df = self._translate_df(translation_vector=translation_vector)
        rotated_df = self._rotate_df(rotation_matrix=rotation_matrix, df=translated_df)
        self.df = self._convert_df_to_cm(conversion_factor=conversion_factor, df=rotated_df)
    
    
    def _identify_duplicates(self)->None:
        # not necessary, since we don't have duplicates in the DLC Tracking
        pass
    
    def _translate_df(self, translation_vector: np.array)->pd.DataFrame:
        """
        Function that translates the raw dataframe to the null space.
        
        Parameter:
            translation_vector(np.Array): vector with offset of xyz to XYZ in each dimension
        Returns:
            translated_df(pandas.DataFrame): the dataframe translated to (0, 0, 0)
        """
        translated_df = self.df_raw.loc[:, ('x', 'y', 'z')] + translation_vector
        return translated_df
    
    def _rotate_df(self, rotation_matrix: Rotation, df: pd.DataFrame)->pd.DataFrame:
        """
        Function, that applies the rotation matrix to the dataframe.
        
        Besides calculating the coordinates, the error is added to the Dataframe.
        
        Parameter:
            rotation_matrix(scipy.spatial.transform.Rotation): Rotation matrix obtained from Euler angles
            df(pandas.DataFrame): the dataframe that will be rotated.
        Returns:
            rotated_df(pandas.DataFrame): the rotated dataframe
        """
        rotated_array = rotation_matrix.apply(df.loc[:, ('x', 'y', 'z')].values, inverse=True)
        rotated_df = pd.DataFrame(rotated_array, columns=['x', 'y', 'z'])
        rotated_df.loc[:, ('x', 'z')] *= -1
        rotated_df['error']=self.df_raw['error']
        return rotated_df
    
    def _convert_df_to_cm(self, conversion_factor: float, df: pd.DataFrame)->pd.DataFrame:
        """
        The coordinates are converted to cm.
        
        Parameters:
            conversion_factor(float): factor to convert the unspecified unit into cm.
            df(pandas.DataFrame): dataframe with unspecified unit
            
        Returns:
            df(pandas.DataFrame): dataframe with values in cm
        """
        df.loc[:, ('x', 'y', 'z')]*=conversion_factor
        return df
        
    def run_basic_operations(self, recorded_framerate: int)->None:
        """
        Function that calculates Speed and Immobility.
        """
        self._get_speed(recorded_framerate = recorded_framerate)
        self._get_rolling_speed()
        self._get_immobility()
        
    def _exclude_frames(self)->None:
        # check for reprojection error (checking for outliers should already be done before triangulation)
        pass
    
    def check_tracking_stability(self, start_end_index: Optional[Tuple]=(0, None))->float:
        """
        Function, that calculates the percentage of frames, in which the marker was detected.
        
        Parameters:
            start_end_index: range in which the percentage of detected labels should be calculated. If no values are passed, the percentage over the total session is returned.
        
        Returns:
            marker_detected_per_total_frames(float)
        """
        marker_detected_per_total_frames = self.df.loc[start_end_index[0]:start_end_index[1], 'x'].count()/self.df.loc[start_end_index[0]:start_end_index[1], :].shape[0]
        return marker_detected_per_total_frames
    
    def _get_speed(self, recorded_framerate: int)->None:
        """
        Function, that calculates the speed of the bodypart, based on the framerate.
        
        After creating an empty column with np.NaN values, the speed is calculated 
        as the squareroot of the squared difference between two frames in -x, -y and -z dimension divided by the duration of a frame.
        
        Parameters:
            recorded_framerate(int): fps of the recording
        """
        self.df.loc[:, 'speed_cm_per_s'] = np.NaN
        self.df.loc[:, 'speed_cm_per_s'] = (np.sqrt(self.df.loc[:, 'x'].diff()**2 + self.df.loc[:, 'y'].diff()**2 + self.df.loc[:, 'z'].diff()**2)) / (1/recorded_framerate)        
    
    
    def _get_rolling_speed(self)->None:
        """
        Function, that applies a sliding window of the size 5 on the speed.
        """
        self.df.loc[:, 'rolling_speed_cm_per_s'] = np.NaN
        self.df.loc[:, 'rolling_speed_cm_per_s'] = self.df.loc[:, 'speed_cm_per_s'].rolling(5, min_periods=3, center=True).mean()

    @property
    def immobility_threshold(self) -> float:
        """ Arbitrary chosen threshold in cm per s for defining immobility."""
        return 1.
    
    def _get_immobility(self)->None:
        """
        Function, that checks frame by frame, whether the rolling_speed of the bodypart is below the immobility threshold.
        """
        self.df.loc[:, 'immobility'] = False
        self.df.loc[self.df['rolling_speed_cm_per_s'] < self.immobility_threshold, 'immobility'] = True     

    def _detect_steps(self)->None:
        """
        Detection of speed peaks based on scipy find_peaks.
        """
        speed = self.df["speed_cm_per_s"].copy()
        #Data smoothening:
        #x = np.arange(0, len(speed))
        #speed = np.nan_to_num(speed, copy=True)
        #spline = interpolate.UnivariateSpline(x, speed, s=1)
        #speed = spline(x)

        peaks = find_peaks(speed, prominence=50)
        steps_per_paw = self._create_steps(steps=peaks[0])
        return steps_per_paw
            
    def _create_steps(self, steps: List)->List['Step']:#as Class Step is not defined yet
        """
        Function, that creates Step objects for every speed peak inside of a gait event.
        
        Parameters:
            List with start_indices for steps.
            
        Returns:
            List with Step elements.
        """
        return [Step(paw = self.id, start_index = step_index) for step_index in steps]
        
                
   

                    
class EventBout():
    """
    Class, that contains start_index, end_index, duration and position of an event.
    
    Attributes:
        self.start_index(int): index of event onset
        self.end_index(int): index of event ending
    """
    def __init__(self, start_index: int, end_index: Optional[int]=0)->None:
        """
        Constructor of class EventBout that sets the attributes start_ and end_index.
        
        Parameters: 
            start_index(int): index of event onset
            end_index(Optional[int]): index of event ending (if event is not only a single frame)
        """
        self.start_index = start_index
        if end_index!=0:
            self.end_index = end_index
        else:
            self.end_index = start_index
        self._create_dict()

    @property
    def freezing_threshold(self) -> float:
        """ Arbitrary chosen threshold in seconds to check for freezing."""
        return 2.

    def check_direction(self, facing_towards_open_end: pd.Series)->None:
        """ 
        Function, that checks the direction of the mouse at the start_index.
        
        Parameters:
            facing_towards_open_end(pandas.Series): Series with boolean values for each frame.
        """
        self.facing_towards_open_end = facing_towards_open_end.iloc[self.start_index]
        self.dict['facing_towards_open_end']=self.facing_towards_open_end

    def check_that_freezing_threshold_was_reached(self, recorded_framerate: int)->None:
        """
        Function, that calculates the duration of an event and checks, whether it exceeded the freezing_threshold.
        
        Parameters:
            recorded_framerate(int): fps of the recording
        """
        self.duration = (self.end_index - self.start_index)/recorded_framerate
        self.freezing_threshold_reached = False
        if self.duration > self.freezing_threshold:
            self.freezing_threshold_reached = True
        self.dict['freezing_threshold_reached']=self.freezing_threshold_reached


    def get_position(self, centerofgravity: Bodypart)->None:
        """
        Function, that saves the position of the mouse at the start_index.
        
        Parameters:
            centerofgravity(Bodypart): object centerofgravity, its df x column is used to extract the mouse position
        """
        self.x_position=centerofgravity.df.loc[self.start_index, 'x']
        self.dict['x_position']=self.x_position

    def _create_dict(self)->None:
        """
        Function that sets the attribut self.dict as Dictionary.
        """
        self.dict = {}
        
        
        
class Angle():
    """
    Class that creates an object for an Angle over a Recording between instances.
    """
    def __init__(self, bodypart_a: Bodypart, bodypart_b: Bodypart, object_to_calculate_angle: Union[np.array, Bodypart])->None:
        """
        Constructor for class Angle.
        
        Depending on the input type of object_to_calculate_angle this class contains functions to calculate the angle of 3 bodyparts to each other at the first given bodypart (bodypart_a)
        if object_to_calculate_angle is type Bodyparr or the angle of 2 bodyparts on a line to a plane if object_to_calculate_angle is a np.array with the plane in coordinate_form.
        
        Parameters:
            bodypart_a(Bodypart): bodypart, at which the angle is calculated.
            bodypart_b(Bodypart): second bodypart, necessary for defining a line
            object_to_calculate_angle(Union): plane as np.array or bodypart necessary for calculating the angle of the line between it and bodypart_a and _b
        
        """
        self.bodypart_a = bodypart_a
        self.bodypart_b = bodypart_b
        if type(object_to_calculate_angle)==Bodypart:
            self.bodypart_c = object_to_calculate_angle
            self.parameter_array = self._calculate_angle_between_three_bodyparts()
        elif type(object_to_calculate_angle)==np.array:
            self.plane = object_to_calculate_angle
            self._calculate_angle_between_bodypart_and_plane()
        
    def _calculate_angle_between_three_bodyparts(self)->np.array:
        """
        Calculates angle at bodypart_a.
        
        After calculating the length of the sides of a triangle ABC, the angles are calculated using law of cosines.
        
        Returns:
            angle(np.array): angle over the whole recording stored in an numpy.Array
        """
        length_a = self._get_length_in_3d_space(self.bodypart_b, self.bodypart_c)
        length_b = self._get_length_in_3d_space(self.bodypart_a, self.bodypart_c)
        length_c = self._get_length_in_3d_space(self.bodypart_a, self.bodypart_b)
        return self._get_angle_from_law_of_cosines(length_a, length_b, length_c)
    
    def _get_length_in_3d_space(self, object_a: Bodypart, object_b: Bodypart) -> np.array:
        """
        Calculates the length between two objects in 3D. 
        
        Parameters:
            object_a(Bodypart)
            object_b(Bodypart)
            
        Returns:
            length(np.array): Length between two objects over the whole recording stored as an numpy.Array.
        """
        if hasattr(object_a, 'df'):       
            length = np.sqrt((object_a.df['x']-object_b.df['x'])**2 + 
                             (object_a.df['y']-object_b.df['y'])**2 + 
                             (object_a.df['z']-object_b.df['z'])**2)
        else:
            length = np.sqrt((object_a.df_raw['x']-object_b.df_raw['x'])**2 + 
                             (object_a.df_raw['y']-object_b.df_raw['y'])**2 + 
                             (object_a.df_raw['z']-object_b.df_raw['z'])**2)
        #theoretisch ist es nicht nötig, den normalisierten df zu nutzen, da der Winkel ja relativ bestimmt wird 
        #und sich daher im df zum raw_df nicht unterscheiden dürfte, dann müsste man allerdings für alle Rechnungen df benutzen, 
        #damit auch die MazeCorners zum rotieren diese Klasse callen können
            
        return length
    
    def _get_angle_from_law_of_cosines(self, length_a: np.array, length_b: np.array, length_c: np.array)->np.array:
        """
        Function that calculates the angle at corner A of a triangle ABC.
        
        After calculating cos(a) using the law of cosines, the inverted cosinus is calculated and it is converted from radians in degrees.
        https://en.wikipedia.org/wiki/Law_of_cosines
        
        Returns: 
            np.array: Angle at corner A in degrees.
        """
        cos_angle = (length_b**2 + length_c**2 - length_a**2) / (2 * length_b * length_c)
        return np.degrees(np.arccos(cos_angle))

    def angle_between_two_lines(self, ax, ay, bx, by, cx, cy, dx, dy):
        #calculates the angle at the intersection of two linear equations, each given by two points (a, b / c, d)
        #following the rule cos(angle)= (m1-m2)/(1+m1*m2)
        m1 = (ay - by) / (ax - bx)
        m2 = (cy - dy) / (cx - dx)
        tan = (m1 - m2) / (1 + m1 * m2)
        angle = np.degrees(np.arctan(tan))
        return angle
    


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_inspect.ipynb.

# %% auto 0
__all__ = ['get_only_matching_xlsx_files', 'get_metadata_from_filename', 'collect_all_available_data', 'filter_dataframe', 'plot',
           'check_data_availability']

# %% ../nbs/02_inspect.ipynb 3
from pathlib import Path, PosixPath
from typing import List, Tuple, Dict, Optional

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

# %% ../nbs/02_inspect.ipynb 6
def get_only_matching_xlsx_files(
    dir_path: Path, paradigm_id: str, week_id: Optional[int] = None
) -> List[Path]:
    filtered_filepaths = []
    for filepath in dir_path.iterdir():
        if filepath.name.endswith(".xlsx") and (paradigm_id in filepath.name):
            if week_id != None:
                if f"week-{week_id}." in filepath.name:
                    filtered_filepaths.append(filepath)
            else:
                filtered_filepaths.append(filepath)
    return filtered_filepaths

# %% ../nbs/02_inspect.ipynb 7
def get_metadata_from_filename(
    filepath_session_results: Path, group_assignment_filepath: Path
) -> Dict:
    metadata = {}
    (
        metadata["line_id"],
        mouse_id,
        metadata["paradigm_id"],
        week_string_with_file_extension,
    ) = filepath_session_results.name.split("_")
    metadata["subject_id"] = f'{metadata["line_id"]}_{mouse_id}'
    metadata["week_id"] = week_string_with_file_extension[
        week_string_with_file_extension.find("-")
        + 1 : week_string_with_file_extension.find(".")
    ]
    df_group_assignment = pd.read_excel(group_assignment_filepath)
    if metadata["subject_id"] in df_group_assignment["subject_id"].unique():
        metadata["group_id"] = df_group_assignment.loc[
            df_group_assignment["subject_id"] == metadata["subject_id"], "group_id"
        ].iloc[0]
    elif (
        metadata["subject_id"] in df_group_assignment["alternative_subject_id"].unique()
    ):
        metadata["group_id"] = df_group_assignment.loc[
            df_group_assignment["alternative_subject_id"] == metadata["subject_id"],
            "group_id",
        ].iloc[0]
    else:
        metadata["group_id"] = "unknown"
    return metadata

# %% ../nbs/02_inspect.ipynb 9
def collect_all_available_data(
    root_dir_path: Path,  # Filepath to the directory that contains the exported results .xlsx files
    group_assignment_filepath: Path,  # Filepath to the group_assignments.xlsx file
    paradigm_ids: List[str],  # List of paradigms of which the results shall be loaded
    week_ids: List[str],  # List of weeks from which the results shall be loaded
    sheet_name: str,  # Tab name of the exported results sheet to load, e.g. "session_overview"
) -> pd.DataFrame:
    all_recording_results_dfs = []
    for week_id in week_ids:
        for paradigm_id in paradigm_ids:
            tmp_matching_filepaths = get_only_matching_xlsx_files(
                dir_path=root_dir_path, paradigm_id=paradigm_id, week_id=week_id
            )
            for filepath in tmp_matching_filepaths:
                metadata = get_metadata_from_filename(
                    filepath_session_results=filepath,
                    group_assignment_filepath=group_assignment_filepath,
                )
                tmp_xlsx = pd.ExcelFile(filepath)
                tmp_df = pd.read_excel(tmp_xlsx, sheet_name=sheet_name, index_col=0)
                for key, value in metadata.items():
                    tmp_df[key] = value
                all_recording_results_dfs.append(tmp_df)
    df = pd.concat(all_recording_results_dfs)
    df.reset_index(drop=True, inplace=True)
    return df

# %% ../nbs/02_inspect.ipynb 13
def filter_dataframe(df: pd.DataFrame, filter_criteria: List[Tuple]) -> pd.DataFrame:
    # assert all list have equal lenghts
    valid_idxs_per_criterion = []
    for column_name, comparison_method, reference_value in filter_criteria:
        # assert valid key in comparison methods
        # assert column name exists
        if comparison_method == "greater":
            valid_idxs_per_criterion.append(
                df.loc[df[column_name] > reference_value].index.values
            )
        elif comparison_method == "smaller":
            valid_idxs_per_criterion.append(
                df.loc[df[column_name] < reference_value].index.values
            )
        elif comparison_method == "equal_to":
            valid_idxs_per_criterion.append(
                df.loc[df[column_name] == reference_value].index.values
            )
        elif comparison_method == "is_in_list":
            valid_idxs_per_criterion.append(
                df.loc[df[column_name].isin(reference_value)].index.values
            )
        elif comparison_method == "is_nan":
            valid_idxs_per_criterion.append(
                df.loc[df[column_name].isnull()].index.values
            )
    shared_valid_idxs_across_all_criteria = valid_idxs_per_criterion[0]
    if len(valid_idxs_per_criterion) > 1:
        for i in range(1, len(valid_idxs_per_criterion)):
            shared_valid_idxs_across_all_criteria = np.intersect1d(
                shared_valid_idxs_across_all_criteria, valid_idxs_per_criterion[i]
            )
    df_filtered = df.loc[shared_valid_idxs_across_all_criteria, :].copy()
    return df_filtered

# %% ../nbs/02_inspect.ipynb 19
def plot(
    df: pd.DataFrame,  # the DataFrame that contains the data you´d like to plot
    x_column: str,  # the column name of the data that should be visualized on the x-axis
    y_column: str,  # the column name of the data that should be visualized on the y-axis
    plot_type: str = "violinplot",  # currently only "violinplot" and "stripplot" are implemented
    hue_column: Optional[
        str
    ] = None,  # if you´d like to use the data of a column to color-code the plotted, you can specify it here (see example below)
    hide_legend: bool = True,  # pass along as `False` if you´d like the legend of the plot to be displayed
) -> None:
    fig = plt.figure(figsize=(8, 5), facecolor="white")
    ax = fig.add_subplot(111)
    if plot_type == "violinplot":
        sns.violinplot(data=df, x=x_column, y=y_column, hue=hue_column)
        if df.shape[0] < 10_000:
            sns.stripplot(
                data=df,
                x=x_column,
                y=y_column,
                hue=hue_column,
                dodge=True,
                color="black",
                alpha=0.3,
            )
    elif plot_type == "stripplot":
        sns.stripplot(data=df, x=x_column, y=y_column, hue=hue_column, dodge=True)
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    if hide_legend:
        ax.get_legend().remove()
    plt.show()

# %% ../nbs/02_inspect.ipynb 22
def check_data_availability(
    root_dir_path: Path, all_week_ids: List[int], all_paradigm_ids: List[str]
) -> pd.DataFrame:
    df = pd.DataFrame({}, index=all_week_ids)
    root_dir_path = root_dir_path
    for week_id in all_week_ids:
        for paradigm_id in all_paradigm_ids:
            all_matching_filepaths = get_only_matching_xlsx_files(
                dir_path=root_dir_path, paradigm_id=paradigm_id, week_id=week_id
            )
            for filepath in all_matching_filepaths:
                subject_id = get_metadata_from_filename(filepath)["subject_id"]
                if f"{subject_id}_{paradigm_id}" not in df.columns:
                    df[f"{subject_id}_{paradigm_id}"] = False
                df.loc[week_id, f"{subject_id}_{paradigm_id}"] = True
    return df
